#include "nmrp.h"
#include "etherboot.h"
#include <nic.h>
#include <asm/rtl865x.h>
#include <rtl_types.h>

int nmrp_packet_recv = 0;
bool nmrp_tftp_wait=false;
bool nmrp_tftp_success=true;
bool nmrp_tftp_runing = false;
extern int power_led_blink_slow; /* add by alfa */
extern int power_led_blink_quick; /* add by alfa */

unsigned long image_dst , image_src ,image_len;

ip_route_t nmrp_server_info;
unsigned char *in_nmrp_pkt;
extern char eth0_mac[6];
extern struct arptable_t  arptable_tftp[3];
extern int flashwrite(unsigned long dst, unsigned long src, unsigned long length);

eth_header_t eth_hdr;
bool nmrp_packet_flag = 0 ;
 
typedef enum {
	NMRPC_NO_ERROR = 0,
	NMRPC_INVALID_PARM
} NMRPC_RET_CODE;


typedef struct {
	int reqCode;
	char reqName[32], respName[32];

} NMRP_REQ_CUSTOM;


#define MIN_ETHER_NMRP_LEN	(ETH_HLEN + NMRP_HDR_LEN)


#define IS_NMRP_PKT(etherNmrpPkt, pktLen, parsedNmrpMsg, sourceMAC) \
	(pktLen > MIN_ETHER_NMRP_LEN &&  \
	(sourceMAC == NULL ||  0 == 0) && \
	NMRP_MsgParsing(etherNmrpPkt, pktLen, parsedNmrpMsg) == NMRP_ERR_NONE)

static bool NMRPConfiguring( char *destMac,bool *fwUpgrade);
static bool NMRPClosing( char *serverMac);

bool NMRPTFTPWaiting(char *destMac);
extern struct nic nic;

int nmrp_status = 0 ;

NMRP_PARSED_OPT *NMRP_MsgGetOpt(NMRP_PARSED_MSG *msg, int optType)
{
	NMRP_PARSED_OPT *opt, *optEnd;
	int i=0;



	for (; i< msg->numOptions; i++)
	{
		if (msg->options[i].type == optType)
			break;
		
	}

	if (i == msg->numOptions || msg->numOptions == 0)
		return 0;
	else 
		return &(msg->options[i]);
}
static bool IsIPAddrValid(unsigned int ipAddr, unsigned int netmask)
{
	unsigned int tmp, mask;

	
	if (ipAddr == 0 || netmask == 0 || (ipAddr & ~netmask) == 0 || (ipAddr & ~netmask) == (0xffffffff & ~netmask))
		return false;

	return true;
}
static int NMRP_OptValParsing(NMRP_PARSED_OPT *optParsed, unsigned char* value)
{
	int retVal = NMRP_ERR_NONE;

	switch (optParsed->type) {
		case NMRP_OPT_MAGIC_NO: /* We require the lenght MUST be correct for the MAGIC-NO option */
			if (optParsed->len != (NMRP_MIN_OPT_LEN + NMRP_MAGIC_NO_LEN) || memcmp(value, NMRP_MAGIC_NO, NMRP_MAGIC_NO_LEN))
				retVal = NMRP_ERR_MSG_INVALID_OPT;
			else
				optParsed->value.streamVal = value;
			break;

		case NMRP_OPT_DEV_IP: /* save the value of IP in the byte stream format, the actual IP is streamVal[0].streamVal[1].streamVal[2].streamVal[3]. */
			//if (  !IsIPAddrValid(ntohl(*(unsigned int *) value), ntohl(*(unsigned int *) (value + 4))) )
			//	retVal = NMRP_ERR_MSG_INVALID_OPT;
			//else
				optParsed->value.streamVal = value;
			break;

		case NMRP_OPT_FW_UP: /* don't check the length of FW-UP option for future compatibility */
			break;

		default:
			retVal = NMRP_ERR_MSG_UNKNOWN_OPT;
			break;
	}

	return retVal;
}

int NMRP_MsgParsing( unsigned char *pkt, int pktLen, NMRP_PARSED_MSG *msg)
{
	if (pktLen < NMRP_HDR_LEN)
		return NMRP_ERR_MSG_INVALID_LEN;

	NMRP_MSG *pkt_nmpr=pkt;
	msg->code = pkt_nmpr->code;
	msg->id = pkt_nmpr->id;
	
	//prom_printf("code = %x id = %x len = %x \n",pkt_nmpr->code, pkt_nmpr->id,pkt_nmpr->length);
	
	if ((msg->length = ntohs(pkt_nmpr->length)) <= pktLen && msg->length >= (NMRP_HDR_LEN)) {
		int retVal;
		int optLen=0;
		NMRP_OPT *opt;
		NMRP_PARSED_OPT *optParsed;

		optParsed = msg->options;
		msg->numOptions = 0;

		pkt +=NMRP_HDR_LEN;
		/* use the length indicated in the NMRP header */
		for (pktLen = msg->length - NMRP_HDR_LEN; pktLen >= NMRP_MIN_OPT_LEN; ) {
			optParsed->type = ntohs(*(unsigned short *)(pkt+optLen));
			optParsed->len = ntohs(*(unsigned short *)(pkt+optLen+2));

			
			if (optParsed->len > pktLen)
				return NMRP_ERR_MSG_INVALID_OPT;

			if(optParsed->len > 4)
			{
			if (((retVal = NMRP_OptValParsing(optParsed, (pkt+optLen+4))) == NMRP_ERR_NONE)) {
				optParsed++;
				if (++msg->numOptions >= NMRP_MAX_OPT_PER_MSG)
					return NMRP_ERR_MSG_TOO_MANY_OPT;
			}
			else if (retVal == NMRP_ERR_MSG_UNKNOWN_OPT)
				return NMRP_ERR_MSG_INVALID_OPT;

			}
			pktLen -= optParsed->len;
			optLen +=optParsed->len;
			optParsed++;
		}
	} /* end if the msg->length is larger than the received packet length */

	//return pktLen == 0 ? NMRP_ERR_NONE : NMRP_ERR_MSG_INVALID_LEN;
	return NMRP_ERR_NONE;
}

bool NMRPRequest( char *destMac, NMRP_REQ_CUSTOM *custom, bool *userData)
{
	int pktLen, retries,retry_start,max_retries;
	//pktbuf_t *pkt;
	NMRP_MSG *msg;
	unsigned char rec_pkt[80];
	//eth_header_t eth_hdr;
	unsigned int configip,configmask;
	int status;
    struct nic rx;
  
	if(custom->reqCode == NMRP_CODE_CONF_REQ )
		max_retries = NMRP_MAX_RETRY_CONF;
	else if (custom->reqCode == NMRP_CODE_CLOSE_REQ )
		max_retries = NMRP_MAX_RETRY_CLOSE;
	
	if(custom->reqCode == NMRP_CODE_TFTP_UL_REQ)
	{
		 if(NULL==(msg=(NMRP_MSG *)malloc(NMRP_HDR_LEN)))
		   { 
		   	nmrp_tftp_success = 0 ;
		    return nmrp_tftp_success ;
		 }
		 msg->reserved = 0;
		 msg->code =custom->reqCode;
		 msg->id = 0;
		 msg->length = htons(NMRP_HDR_LEN);
		 prepare_txpkt(0,ETHER_NMRP,destMac,(Int8*)msg,NMRP_HDR_LEN);
		 free(msg);
		 
		 nmrp_tftp_wait=true;
		 nmrp_tftp_runing = true;
		
		 tftpd_entry();
		 
         if(nmrp_tftp_success)
            prom_printf("NMRP TFTP Server success \n");
         return nmrp_tftp_success;
	
	}
 	
    if(NULL==(msg=(NMRP_MSG *)malloc(NMRP_HDR_LEN)))
			return -1 ;
	msg->reserved = 0;
	msg->code = custom->reqCode;
	msg->id = 0;
	msg->length = htons(NMRP_HDR_LEN); /* header only, no option */
	retries = 0;

	/* send the packet */
		
	do{
	    prepare_txpkt(0,ETHER_NMRP,destMac,(Int8*)msg,NMRP_HDR_LEN);	
		retry_start = get_timer_jiffies();
		do{
			pktLen = 0;
			 
            if((status=polling_rx()) == 1)
 	          {
 	           rx.packet = mac_rx_buff[buff_rd_index].packet;
 	           rx.packetlen = mac_rx_buff[buff_rd_index].packet_size;
               ++buff_rd_index;
               if(buff_rd_index > ( MAX_RX_BUFF-1 ))
                  buff_rd_index = 0 ;
               kick_tftpd(rx); 
               buff_bz_num--;
                
               if(nmrp_packet_flag == 1)
			     {
			       pktLen = rx.packetlen;
			       nmrp_packet_flag = 0;
			      }  	
			   }
		
			if ( pktLen > 0 && memcmp(&(eth_hdr.source),destMac,6) ==0 && memcmp(&(eth_hdr.destination),eth0_mac,6)==0)
			  {
				
				NMRP_MSG 	*nmrp_pkt=(NMRP_MSG *)in_nmrp_pkt;
				int opt_len=nmrp_pkt->length - NMRP_HDR_LEN;
				NMRP_PARSED_OPT *pase_opt;
				if(custom->reqCode==NMRP_CODE_CONF_REQ )
				 {	
				  if(nmrp_pkt->code == NMRP_CODE_CONF_ACK)
					{
						prom_printf("NMRP Config Ack received \n"); 
						pase_opt = (NMRP_PARSED_OPT *)(in_nmrp_pkt+NMRP_HDR_LEN);
						if(pase_opt->type == NMRP_OPT_DEV_IP)
						{
						//	prom_printf("NMRP_OPT_DEV_IP\n"); 
							memcpy(&configip,(in_nmrp_pkt+NMRP_HDR_LEN +4),4);
							memcpy(&configmask,(in_nmrp_pkt+NMRP_HDR_LEN +8),4);
							if(IsIPAddrValid(configip,configmask))
								memcpy(&(arptable_tftp[TFTP_SERVER].ipaddr.s_addr), (in_nmrp_pkt+NMRP_HDR_LEN +4), 4);
							else 
								continue;
						}
						else
							continue;
						opt_len -=pase_opt->len;
						if(opt_len >= 4 && (*(unsigned short *)(in_nmrp_pkt+NMRP_HDR_LEN+pase_opt->len))==NMRP_OPT_FW_UP )
							*userData = true;
						free(msg);
			 			return true;				 
					}
					else
						continue;
				 }	
				else if (custom->reqCode==NMRP_CODE_CLOSE_REQ)
				 {
					if(nmrp_pkt->code == NMRP_CODE_CLOSE_ACK)
					{
						free(msg);
						return true;
					}
					else
						continue;
				 }					
			 }
		}while ( (get_timer_jiffies() - retry_start) < 50  );  // 0.5 sec
	} while ((++retries) <= max_retries);
	free(msg);
	return false;
}

static bool NMRPConfiguring( char *destMac, bool *fwUpgrade)
{
	NMRP_REQ_CUSTOM reqConf = { NMRP_CODE_CONF_REQ, "CONF-REQ", "CONF-ACK"};

	if (NMRPRequest(destMac, &reqConf, fwUpgrade) == false)
		return false;

	return true;
}

bool NMRPTFTPWaiting(char *destMac)
{
        NMRP_REQ_CUSTOM reqTftpUl = { NMRP_CODE_TFTP_UL_REQ, "TFTP-UL-REQ", "TFTP-WRQ"};

        return NMRPRequest(destMac, &reqTftpUl, NULL);
}

static bool NMRPClosing(char *serverMac)
{
	NMRP_REQ_CUSTOM reqClose = { NMRP_CODE_CLOSE_REQ, "CLOSE-REQ", "CLOSE-ACK"};

	return NMRPRequest( serverMac, &reqClose, NULL);
}

void nmrp_packet(struct nic rx)
{
  in_nmrp_pkt = (unsigned char *)&rx.packet[ETH_HLEN];
  memcpy(&(eth_hdr.destination),&rx.packet[0],sizeof(eth_header_t));
  nmrp_packet_flag = 1;
}

extern volatile int get_timer_jiffies(void);

bool nmrp_entry(void)
{
 bool succeeds,fwUpgrade;
 int listening_start,packet_wait;
 NMRP_PARSED_MSG nmrpMsg;
 int pktLen,status;
 struct nic rx;
 bool listening = true ;
 bool ipset = false;
 succeeds = false;
  
 while(!succeeds) 
	{
		prom_printf("\n");
     if (listening) 
		{ //##### Listening state #####
		 prom_printf("NMRP Client starts Listening for ADVERTISE...\n"); 
		 listening_start=get_timer_jiffies();
		 do{	
		    pktLen = 0;
		    if((status=polling_rx()) == 1)
 	          {
 	       	   rx.packet = mac_rx_buff[buff_rd_index].packet;
 	           rx.packetlen = mac_rx_buff[buff_rd_index].packet_size;
               ++buff_rd_index;
               if(buff_rd_index > ( MAX_RX_BUFF-1 ))
                  buff_rd_index = 0 ;
               kick_tftpd(rx); 
               buff_bz_num--;
                
               if(nmrp_packet_flag == 1)
			     {
			      pktLen = rx.packetlen;
			      nmrp_packet_flag = 0;
		         }  	
               }
     
			if ( pktLen > 0 )
				{
					if (IS_NMRP_PKT(in_nmrp_pkt, pktLen, &nmrpMsg, NULL) && nmrpMsg.code == NMRP_CODE_ADVERTISE &&  NMRP_MsgGetOpt(&nmrpMsg, NMRP_OPT_MAGIC_NO) != NULL)
					{
						listening = false;
						break;
					}
				}
		    }while ( (get_timer_jiffies() - listening_start) < 400  );  // 4 sec
			if (listening)
			{
				prom_printf("NMRP timeoust exit;");
				return false;
			}
			else
			{
				ipset = false;
				memcpy(&(nmrp_server_info.enet_addr),&(eth_hdr.source),sizeof(enet_addr_t));
				prom_printf("NMRP advertise received.\n");
			}
		  }	
		   if ( !ipset ) { //##### Configuring state #####
			prom_printf("NMRP client Configuring Request send\n");

			if (NMRPConfiguring(&nmrp_server_info.enet_addr, &fwUpgrade) == false) {
				listening = true;
				prom_printf("timeouts...Back to Listening state.\n");
				continue;
			}

			prom_printf("NMRP IP setting success\n");
			ipset = true;
		}
		if (fwUpgrade) 
		{

			prom_printf("NMRP Client TFTP Upload Request send\n");
			nmrp_tftp_success = true;
			
			if (NMRPTFTPWaiting(&(nmrp_server_info.enet_addr)) == false) { //##### TFTP-Waiting state #####
				prom_printf("timeouts...Back to tftp waiting state.\n");
				continue;
			}
			
            prom_printf("NMRP Client Close Request send\n");
			if (NMRPClosing(&(nmrp_server_info.enet_addr)) == false)
				prom_printf("NMRP timeouts...exits without CLOSE-ACK.\n");
			else
				prom_printf("NMRP received CLOSE-ACK.\nNMRP Client done, exits.\n");
		    
		    if (flashwrite(image_dst, image_src, image_len))
		       {
		       	 prom_printf("\nNMRP write firmware to flash finished\n");	
		       	 power_led_blink_slow = 0;
		       	 power_led_blink_quick = 1;
		       	 while(1);
		       }	 
		    else
		        {
		         prom_printf("NMRP write firmware to flash fail\n");	
		         nmrp_tftp_wait=false;
                 nmrp_tftp_success=true;
                 nmrp_tftp_runing = false;	  
                 continue;
                } 
			nmrp_tftp_wait = false;		
			succeeds = true;
			
		} 
	
	}
}			


