/*
 *  Boa, an http server
 *  Copyright (C) 1995 Paul Phillips <psp@well.com>
 *  Some changes Copyright (C) 1996,97 Larry Doolittle <ldoolitt@jlab.org>
 *  Some changes Copyright (C) 1997,99 Jon Nelson <jnelson@boa.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 1, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/* $Id: read.c,v 1.44 2001/10/20 02:58:39 jnelson Exp $*/

#include "boa.h"

/*
 * Name: read_header
 * Description: Reads data from a request socket.  Manages the current
 * status via a state machine.  Changes status from READ_HEADER to
 * READ_BODY or WRITE as necessary.
 *
 * Return values:
 *  -1: request blocked, move to blocked queue
 *   0: request done, close it down
 *   1: more to do, leave on ready list
 */

#if WAN_LAN_CONFLICT

int hijack = 0;
char hijackdomain[64];
char hijackurl[256];

void debug_echo(char * str)
{
    char buff[256] = "";

    sprintf(buff,"echo %s >/dev/console",str);
    system(buff);
}

char *get_lan_ip()
{
		return nvram_safe_get("lan_ipaddr");
}

int open_hijack_conflict(char *filepath)
{
   FILE *fp = NULL;
   int val = 0;
   char *ptr;

#if 1
   ptr = nvram_get(filepath);
   if(!ptr) {
       hijack = 0;
       return 0;
   }
   val = atoi(ptr);
   if(strcmp(filepath,DNS_HIJACK) == 0) {
		hijack = val;
	}
 
   return val;
#else
   if(!(fp = fopen(filepath, "r"))) {
       //debug_echo(filepath);
       hijack = 0;
		return 0;
   }

   val = getc(fp) - '0';
   fclose(fp);

   if(strcmp(filepath,DNS_HIJACK) == 0) {
		hijack = val;
	}
   return val;
#endif
}

int save_req_domain(char *domian)
{
    FILE *fp;

    if(!domian)
        return 0;
    if(!(fp = fopen("/tmp/hijack_domain", "w")))
        return 0;
    fprintf(fp,"%s", domian);
    fclose(fp);
    return 0;
}

int save_req_url(char *url)
{
    FILE *fp;

    if(!url)
        return 0;
    if(!(fp = fopen("/tmp/hijack_url", "w")))
        return 0;
    fprintf(fp,"%s", url);
    fclose(fp);
    return 0;
}

/*
 * returns true if the string passed in is an internet address in dotted quad
 * notation.
 */
int is_dotted_quad(char *addr)
{
  int q[4];
  char *p;
  int i;

  if(sscanf(addr, "%d.%d.%d.%d", &(q[0]), &(q[1]), &(q[2]), &(q[3])) != 4)
  {
    return(0);
  }

  if(q[0] > 255 || q[0] < 0 ||
      q[1] > 255 || q[1] < 0 ||
      q[2] > 255 || q[2] < 0 ||
      q[3] > 255 || q[3] < 0)
  {
    return(0);
  }

  /* we know there are 3 dots */
  p = addr;
  if(p != NULL) { p = strchr(p, '.'); p++; }
  if(p != NULL) { p = strchr(p, '.'); p++; }
  if(p != NULL) { p = strchr(p, '.'); }
  for(i=0; *p != '\0' && i<4; i++, p++);
  if(*p != '\0')
  {
    return(0);
  }

  return(1);
}

int is_url_local(char *url)
{
	char pathname[64];

	if( url[0] != '/')
		return 0;
	if(strlen(url) >= 64)
		return 0;
	if(strcmp(url,"/") == 0)
		return 0;

	if(strncmp(url,"/html/", 6) == 0) {
		sprintf(pathname,"/www%s", url);
		if(access(pathname, F_OK) < 0) {
			return 0;
		}
	} else if(strncmp(url,"/cgi-bin/",9) == 0) {
        sprintf(pathname,"/www%s",url );
        if(access(pathname, F_OK) < 0) {
            return 0;
        }
    } else {
		sprintf(pathname,"/www/html%s",url );
		if(access(pathname, F_OK) < 0) {
			sprintf(pathname,"/www/cgi-bin%s",url );
			if(access(pathname, F_OK) < 0) {
				return 0;
			}
		}
	}
	return 1;
}

int get_req_domian(request *req)
{
	char *check = NULL, *buffer = NULL, *stop = NULL, *host = NULL,*url = NULL, *filepath = NULL;
	int bytes;
	int ret = 0;

	check = req->client_stream + req->parse_pos;
	buffer = req->client_stream;
	bytes = req->client_stream_pos;

	if (memcmp(req->client_stream, "GET ", 4))
		return 0;

	while (check < (buffer + bytes)) {
		if(*check == 'H') {
			if(memcmp(check, "Host: ", 6) != 0) {
				check++;
				continue;
			}
			check += 3;
			if(*check != ' ')
				++check;
			while(*(++check) == ' ');

			stop = check;
			while(*stop != '\r' && *stop != '\n')
				stop ++;
			if((stop == check) || (stop - check > MAX_HEADER_LENGTH))
				return 0;

			if(!(host = (char *)malloc(stop - check + 1)))
				return 0;

			memcpy(host, check, stop - check);
			host[stop - check] = '\0';

			if(strstr(host,"members.dyndns.org")) {
				free(host);
				return 0;
			}

			if(strstr(host, "routerlogin.com") || strstr(host, "routerlogin.net")) {
				free(host);
                return 0;
            }

			if(is_dotted_quad(host)){
				free(host);
				return 0;
			}

			save_req_domain(host);
			ret = 1;
			break;
		}
		check++;
	}

	if(ret) {
    	url = req->client_stream + 3;
    	if (*url != ' ')
        	++url;

		while (*(++url) == ' ');

    	stop = url;
    	while (*stop != ' ' && *stop != '\r' && *stop != '\n')
        	++stop;
    	if (stop - url > MAX_HEADER_LENGTH) {
        	free(host);
			return 0;
    	}
    	if(!(filepath = (char *)malloc(stop - url + 1))) {
        	free(host);
			return 0;
    	}
    	memcpy(filepath, url, stop - url);
    	filepath[stop - url] = '\0';
		if(is_url_local(filepath) == 1){
			free(host);
			free(filepath);
			return 0;
		}
		if( host && strncmp(hijackdomain, host, sizeof(hijackdomain)) != 0) {
			strncpy(hijackdomain, host, sizeof(hijackdomain));
			strncpy(hijackurl, filepath, sizeof(hijackurl));
			save_req_url(filepath);
        }
    	if(hijack)
			send_r_hijack(req);
	/*	else {
			if (open_hijack_conflict(DNS_HIJACK) != 1)
			{
				send_r_wanlan_conflict(req);
			}
		}*/
		req->status = DONE;
		free(host);
		free(filepath);
	}

	return ret;
}

#endif
int read_header(request * req)
{
    int bytes, buf_bytes_left;
    char *check, *buffer;

    check = req->client_stream + req->parse_pos;
    buffer = req->client_stream;
    bytes = req->client_stream_pos;

#ifdef VERY_FASCIST_LOGGING
    if (check < (buffer + bytes)) {
        buffer[bytes] = '\0';
        log_error_time();
        fprintf(stderr, "%s:%d - Parsing headers (\"%s\")\n",
                __FILE__, __LINE__, check);
    }
#endif


#if WAN_LAN_CONFLICT
	if(open_hijack_conflict(DNS_HIJACK)) {
		if(get_req_domian(req) == 1) {
			return 0;
		 }
	}
#endif

    while (check < (buffer + bytes)) {
        switch (req->status) {
        case READ_HEADER:
            if (*check == '\r') {
                req->status = ONE_CR;
                req->header_end = check;
            } else if (*check == '\n') {
                req->status = ONE_LF;
                req->header_end = check;
            }
            break;

        case ONE_CR:
            if (*check == '\n')
                req->status = ONE_LF;
            else if (*check != '\r')
                req->status = READ_HEADER;
            break;

        case ONE_LF:
            /* if here, we've found the end (for sure) of a header */
            if (*check == '\r') /* could be end o headers */
                req->status = TWO_CR;
            else if (*check == '\n')
                req->status = BODY_READ;
            else
                req->status = READ_HEADER;
            break;

        case TWO_CR:
            if (*check == '\n')
                req->status = BODY_READ;
            else if (*check != '\r')
                req->status = READ_HEADER;
            break;

        default:
            break;
        }

#ifdef VERY_FASCIST_LOGGING
        log_error_time();
        fprintf(stderr, "status, check: %d, %d\n", req->status, *check);
#endif

        req->parse_pos++;       /* update parse position */
        check++;

        if (req->status == ONE_LF) {
            *req->header_end = '\0';

            /* terminate string that begins at req->header_line */

            if (req->logline)
                process_option_line(req);
            else {
                if (process_logline(req) == 0)
                    return 0;
                if (req->simple)
                    return process_header_end(req);
            }
            /* set header_line to point to beginning of new header */
            req->header_line = check;
        } else if (req->status == BODY_READ) {
#ifdef SOAP_ON
	if (req->method == M_POST && req->soap_action) 
		return ExecuteSoapAction(req, 
			req->soap_action, req->soap_actionlen);
#endif	
#ifdef VERY_FASCIST_LOGGING
            int retval;
            log_error_time();
            fprintf(stderr, "%s:%d -- got to body read.\n",
                    __FILE__, __LINE__);
            retval = process_header_end(req);
#else
            int retval = process_header_end(req);
#endif
            /* process_header_end inits non-POST cgi's */

            if (retval && req->method == M_POST) {
                /* for body_{read,write}, set header_line to start of data,
                   and header_end to end of data */
                req->header_line = check;
                req->header_end =
                    req->client_stream + req->client_stream_pos;

                req->status = BODY_WRITE;
                /* so write it */
                /* have to write first, or read will be confused
                 * because of the special case where the
                 * filesize is less than we have already read.
                 */

                /*

                   As quoted from RFC1945:

                   A valid Content-Length is required on all HTTP/1.0 POST requests. An
                   HTTP/1.0 server should respond with a 400 (bad request) message if it
                   cannot determine the length of the request message's content.

                 */

                if (req->content_length) {
                    req->filesize = atoi(req->content_length);
                    req->filepos = 0;
		     if(req->filesize > 1024*1000){
			//    system("/usr/bin/killall upnp"); system("/usr/bin/killall ripd");
			  //  system("/usr/bin/killall zebra"); system("/usr/bin/killall syslogd");
			   // system("/usr/bin/killall dnsmasq");
			   sleep(2);
		     }
                    if (single_post_limit && req->filesize > single_post_limit) {
                        log_error_time();
                        fprintf(stderr, "Content-Length [%ld] > SinglePostLimit [%d] on POST!\n",
                               req->filesize, single_post_limit);
                        send_r_bad_request(req);
                        return 0;
                    }
                } else {
                    log_error_time();
                    fprintf(stderr, "Unknown Content-Length POST!\n");
                    send_r_bad_request(req);
                    return 0;
                }
                if (req->header_end - req->header_line > req->filesize) {
                    req->header_end = req->header_line + req->filesize;
                }
            }                   /* either process_header_end failed or req->method != POST */
            return retval;      /* 0 - close it done, 1 - keep on ready */
        }                       /* req->status == BODY_READ */
    }                           /* done processing available buffer */

#ifdef VERY_FASCIST_LOGGING
    log_error_time();
    fprintf(stderr, "%s:%d - Done processing buffer.  Status: %d\n",
            __FILE__, __LINE__, req->status);
#endif

    if (req->status < BODY_READ) {
        /* only reached if request is split across more than one packet */

        buf_bytes_left = CLIENT_STREAM_SIZE - req->client_stream_pos;
        if (buf_bytes_left < 1) {
            log_error_time();
            fputs("buffer overrun - read.c, read_header - closing\n",
                  stderr);
            req->status = DEAD;
            return 0;
        }

        bytes = read(req->fd, buffer + req->client_stream_pos, buf_bytes_left);

        if (bytes < 0) {
            if (errno == EINTR)
                return 1;
            if (errno == EAGAIN || errno == EWOULDBLOCK) /* request blocked */
                return -1;
            /*
               else if (errno == EBADF || errno == EPIPE) {

               req->status = DEAD;
               return 0;
             */
            boa_perror(req, "header read");
            return 0;
        } else if (bytes == 0) {
            /*
               log_error_time();
               fputs("unexpected end of headers\n", stderr);
             */
            return 0;
        }

        /* bytes is positive */
        req->client_stream_pos += bytes;

#ifdef FASCIST_LOGGING1
        log_error_time();
        req->client_stream[req->client_stream_pos] = '\0';
        fprintf(stderr, "%s:%d -- We read %d bytes: \"%s\"\n",
                __FILE__, __LINE__, bytes,
#ifdef VERY_FASCIST_LOGGING2
                req->client_stream + req->client_stream_pos - bytes);
#else
                "");
#endif
#endif

        return 1;
    }
    return 1;
}

/*
 * Name: read_body
 * Description: Reads body from a request socket for POST CGI
 *
 * Return values:
 *
 *  -1: request blocked, move to blocked queue
 *   0: request done, close it down
 *   1: more to do, leave on ready list
 *

 As quoted from RFC1945:

 A valid Content-Length is required on all HTTP/1.0 POST requests. An
 HTTP/1.0 server should respond with a 400 (bad request) message if it
 cannot determine the length of the request message's content.

 */

int read_body(request * req)
{
    int bytes_read, bytes_to_read, bytes_free;

    bytes_free = BUFFER_SIZE - (req->header_end - req->header_line);
    bytes_to_read = req->filesize - req->filepos;

    if (bytes_to_read > bytes_free)
        bytes_to_read = bytes_free;

    if (bytes_to_read <= 0) {
        req->status = BODY_WRITE; /* go write it */
        return 1;
    }

    bytes_read = read(req->fd, req->header_end, bytes_to_read);

    if (bytes_read == -1) {
        if (errno == EWOULDBLOCK || errno == EAGAIN) {
            /*
               req->status = BODY_WRITE;
               return 1;
             */
            return -1;
        } else {
            boa_perror(req, "read body");
            return 0;
        }
    } else if (bytes_read == 0) {
        /* this is an error.  premature end of body! */
        log_error_time();
        fprintf(stderr, "%s:%d - Premature end of body!!\n",
                __FILE__, __LINE__);
        send_r_bad_request(req);
        return 0;
    }

    req->status = BODY_WRITE;

#ifdef FASCIST_LOGGING1
    log_error_time();
    fprintf(stderr, "%s:%d - read %d bytes.\n",
            __FILE__, __LINE__, bytes_to_read);
#endif

    req->header_end += bytes_read;

    return 1;
}

/*
 * Name: write_body
 * Description: Writes a chunk of data to a file
 *
 * Return values:
 *  -1: request blocked, move to blocked queue
 *   0: EOF or error, close it down
 *   1: successful write, recycle in ready queue
 */

int write_body(request * req)
{
    int bytes_written, bytes_to_write = req->header_end - req->header_line;
    if (req->filepos + bytes_to_write > req->filesize)
        bytes_to_write = req->filesize - req->filepos;

    if (bytes_to_write == 0) {  /* nothing left in buffer to write */
        req->header_line = req->header_end = req->buffer;
        if (req->filepos >= req->filesize)
            return init_cgi(req);
        /* if here, we can safely assume that there is more to read */
        req->status = BODY_READ;
        return 1;
    }
    bytes_written = write(req->post_data_fd,
                          req->header_line, bytes_to_write);

    if (bytes_written == -1) {
        if (errno == EWOULDBLOCK || errno == EAGAIN)
            return -1;          /* request blocked at the pipe level, but keep going */
        else if (errno == EINTR)
            return 1;
        else if (errno == ENOSPC) {
            /* 20010520 - Alfred Fluckiger */
            /* No test was originally done in this case, which might  */
            /* lead to a "no space left on device" error.             */
            boa_perror(req, "write body"); /* OK to disable if your logs get too big */
            return 0;
        } else {
            boa_perror(req, "write body"); /* OK to disable if your logs get too big */
            return 0;
        }
    }
#ifdef FASCIST_LOGGING
    log_error_time();
    fprintf(stderr, "%s:%d - wrote %d bytes. %ld of %ld\n",
            __FILE__, __LINE__,
            bytes_written, req->filepos, req->filesize);
#endif

    req->filepos += bytes_written;
    req->header_line += bytes_written;

    return 1;                   /* more to do */
}
