/*
 * FILE NAME rtl_gpio.c
 *
 * BRIEF MODULE DESCRIPTION
 *  GPIO For Flash Reload Default
 *
 *  Author: jimmylin@realtek.com.tw
 *
 * Copyright 2005 Realtek Semiconductor Corp.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE	LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */



#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <asm/errno.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/miscdevice.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/reboot.h>
#include <linux/kmod.h>
#include <linux/proc_fs.h>

#define AUTO_CONFIG

// 2009-0414
//#define	DET_WPS_SPEC
#ifdef CONFIG_RTL_EB8186
	#define READ_RF_SWITCH_GPIO
#endif

#ifdef CONFIG_RTL8196B
	#define READ_RF_SWITCH_GPIO
#endif

#if defined(CONFIG_RTL865X)
	#include <asm/rtl865x/rtl865xc_asicregs.h>

	#define RTL_GPIO_PABDIR		PABCD_DIR
	#define RTL_GPIO_PABDATA	PABCD_DAT
	#define RTL_GPIO_PABCNR		PABCD_CNR
	#define AUTOCFG_BTN_PIN		15
	#define AUTOCFG_LED_PIN		14
	#define RESET_LED_PIN		13
	#define RESET_BTN_PIN		15
#endif
#if defined(CONFIG_RTL8196B)
	#include <asm/rtl865x/rtl865xc_asicregs.h>
	#define RTL_GPIO_PABDIR		PABCD_DIR
	#define RTL_GPIO_PABDATA	PABCD_DAT
	#define RTL_GPIO_PABCNR		PABCD_CNR
	#define AUTOCFG_BTN_PIN		1
	#define AUTOCFG_LED_PIN		20
	#define RESET_LED_PIN		18
	#define RESET_BTN_PIN		0
	#define RTL_GPIO_MUX 0xB8000030
#if defined(CONFIG_DNI_JTAG_ON)
	#define RTL_GPIO_MUX_DATA 0x0FC00080//for WIFI ON/OFF and GPIO
#else
	#define RTL_GPIO_MUX_DATA 0x0FC00380//for WIFI ON/OFF and GPIO
#endif
	#define RTL_GPIO_WIFI_ONOFF	19
#if defined(CONFIG_DNI_LED)
	#define DNI_POWER_LED	17
#endif


#endif


#ifdef CONFIG_RTL_EB8186
	#define RTL_GPIO_PABDIR		0xbd010124
	#define RTL_GPIO_PABDATA	0xbd010120
	#define RESET_BTN_PIN		0
	#define AUTOCFG_BTN_PIN		1
	#define RESET_LED_PIN		2
	#define AUTOCFG_LED_PIN		3
	#define AUTOCFG_LED_OFF		0
	#define AUTOCFG_LED_ON		1
	#define AUTOCFG_LED_BLINK	2
#if 0
#else
	#define RTL_GPIO_PABDIR		0xbd010040
	#define RTL_GPIO_PABDATA	0xbd010044
	#define RTL_WLAN_CR9346		0xbd400050
	#define RTL_WLAN_CONFIG0	0xbd400051
	#define RTL_WLAN_PSR		0xbd40005e
	#define RESET_BTN_PIN		10
#endif
#endif

#ifdef CONFIG_RTL865X_KLD
	#define PS_LED_GREEN_PIN	13
	#define PS_LED_ORANGE_PIN	14
	#define INET_LED_GREEN_PIN	24
	#define INET_LED_ORANGE_PIN	25
	#define GPIO_IRQ_NUM		1	
#endif

#ifdef CONFIG_RTL8196B_KLD
	#define PS_LED_GREEN_PIN	18
	#define PS_LED_ORANGE_PIN	3
	#define INET_LED_GREEN_PIN	8
	#define INET_LED_ORANGE_PIN	17
	#define GPIO_IRQ_NUM		1	
	static char test_mode_flag = '0';
	static unsigned int    rst_btn_flag;
	static unsigned int    wps_btn_flag;
#endif

// 2009-0414
#ifdef DET_WPS_SPEC
	#define GPIO_IRQ_NUM		1	
#endif
#ifdef CONFIG_RTL8186_KB
	#define PROBE_TIME	10
#elif defined(CONFIG_RTL865X_SC)
	#define PROBE_TIME	10
#elif defined(CONFIG_DNI_LED)
	#define TICKS_PER_SEC	100
	#define RTL_GPIO_TIMER_RESOLUTION		10  //10 Ticks, 0.1 second
	//0~2:do nothing, 2~5: reboot, 5~ :reset_default
	#define REBOOT_TIME	(2 * TICKS_PER_SEC / RTL_GPIO_TIMER_RESOLUTION)  
	#define PROBE_TIME	(5 * TICKS_PER_SEC / RTL_GPIO_TIMER_RESOLUTION)  
#else
	#define PROBE_TIME	5
#endif

#define PROBE_NULL		0
#define PROBE_ACTIVE	1
#define PROBE_RESET		2
#define PROBE_RELOAD	3
#define RTL_R32(addr)		(*(volatile unsigned long *)(addr))
#define RTL_W32(addr, l)	((*(volatile unsigned long*)(addr)) = (l))
#define RTL_R8(addr)		(*(volatile unsigned char*)(addr))
#define RTL_W8(addr, l)		((*(volatile unsigned char*)(addr)) = (l))

//#define  GPIO_DEBUG
#ifdef GPIO_DEBUG
/* note: prints function name for you */
#  define DPRINTK(fmt, args...) printk("%s: " fmt, __FUNCTION__ , ## args)
#else
#  define DPRINTK(fmt, args...)
#endif

static struct timer_list probe_timer;
static unsigned int    probe_counter;
static unsigned int    probe_state;

static char default_flag='0';
//Brad add for update flash check 20080711
#if !defined(CONFIG_RTL8186_TR) && !defined(CONFIG_RTL8196B_TR) && !defined(CONFIG_RTL865X_AC) && !defined(CONFIG_RTL865X_KLD) && !defined(CONFIG_RTL8196B_KLD) && !defined(CONFIG_RTL865X_SC)
int start_count_time=0;
int Reboot_Wait=0;
#endif
#ifndef CONFIG_RTL_EB8186
unsigned int led0enable;
#endif

#ifdef CONFIG_RTL8186_KB
int	disable_power_led_blink=0;
#endif

#if defined(CONFIG_RTL8186_TR) || defined(CONFIG_RTL8196B_TR)
int	system_led_blink=0;
extern int flash_write_flag;
#if 0 //Brad disable, since the check is not precise for system status			
unsigned int sys_keep_alive = 0;
#endif
#endif

//#ifdef CONFIG_RTL865X_AC
#if defined(CONFIG_RTL865X_AC) || defined(CONFIG_RTL865X_KLD) || defined(CONFIG_RTL8196B_KLD)
int	system_led_blink=0;
extern int flash_write_flag;
#if 0 //Brad disable, since the check is not precise for system status			
unsigned int sys_keep_alive = 0;
#endif
int kld_model=0;
#endif

#if defined(CONFIG_RTL865X_KLD) || defined(CONFIG_RTL8196B_KLD)
static char ps_led_flag = 0;
static struct timer_list ps_led_timer;
static char ps_green_toggle = 0;
static char ps_orange_toggle = 0;

static char inet_led_flag = 0;
static struct timer_list inet_led_timer;
static char inet_green_toggle = 0;
static char inet_orange_toggle = 0;
static int wps_button_push = 0;

static struct timer_list wps_QuickBlink_timer; 
static void ps_led_process(void);
#endif

#ifdef	DET_WPS_SPEC
static int wps_button_push = 0;

#endif

#ifdef AUTO_CONFIG
static unsigned int		AutoCfg_LED_Blink;
static unsigned int		AutoCfg_LED_Toggle;
#if (CONFIG_DNI_LED)
#if 0
static unsigned int		AutoCfg_LED_Blink_Time;
static unsigned int		AutoCfg_LED_Blink_Timeout;
#endif
static unsigned int		AutoCfg_LED_ON_MAx_Count;
static unsigned int		AutoCfg_LED_OFF_MAx_Count;
static unsigned int		AutoCfg_LED_ON_Count;
static unsigned int		AutoCfg_LED_OFF_Count;
#endif

#if defined(CONFIG_RTL865X_SC)
static unsigned int ResetToAutoCfgBtn = 0;
#endif
#ifdef CONFIG_RTL865X_KLD
void ps_led_gpio_init(void)
{
	if(kld_model==1){
	RTL_W32(PABCD_CNR, (RTL_R32(PABCD_CNR) & (~(1 << PS_LED_GREEN_PIN))));
	RTL_W32(PABCD_DIR, (RTL_R32(PABCD_DIR) | (1 << PS_LED_GREEN_PIN)));
	RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << PS_LED_GREEN_PIN)));
	RTL_W32(PEFGH_CNR, (RTL_R32(PEFGH_CNR) & (~((1 << PS_LED_ORANGE_PIN) | (1 << INET_LED_GREEN_PIN) | (1 << INET_LED_ORANGE_PIN)))));
	RTL_W32(PEFGH_DIR, (RTL_R32(PEFGH_DIR) | ((1 << PS_LED_ORANGE_PIN) | (1 << INET_LED_GREEN_PIN) | (1 << INET_LED_ORANGE_PIN))));
	RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) & (~(1 << PS_LED_ORANGE_PIN))));
	RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) | ((1 << INET_LED_GREEN_PIN) | (1 << INET_LED_ORANGE_PIN))));
	}else{

	RTL_W32(PABCD_CNR, (RTL_R32(PABCD_CNR) & (~(1 << PS_LED_GREEN_PIN))));
	RTL_W32(PABCD_DIR, (RTL_R32(PABCD_DIR) | (1 << PS_LED_GREEN_PIN)));
	RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << PS_LED_GREEN_PIN)));
	}
	
}
#endif
#ifdef CONFIG_RTL8196B_KLD
void ps_led_gpio_init(void)
{
	RTL_W32(0xB8000030, (RTL_R32(0xB8000030) | 0x00F00F80 ));
	
	if(kld_model==1){
	/* ps_led init setting */
	RTL_W32(PABCD_CNR, (RTL_R32(PABCD_CNR) & (~(1 << PS_LED_GREEN_PIN))));
	RTL_W32(PABCD_CNR, (RTL_R32(PABCD_CNR) & (~(1 << PS_LED_ORANGE_PIN))));

	RTL_W32(PABCD_DIR, (RTL_R32(PABCD_DIR) | (1 << PS_LED_GREEN_PIN)));
	RTL_W32(PABCD_DIR, (RTL_R32(PABCD_DIR) | (1 << PS_LED_ORANGE_PIN)));
	
	RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << PS_LED_GREEN_PIN)));
	RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << PS_LED_ORANGE_PIN))));

	/* inet_led init setting */
	RTL_W32(PABCD_CNR, (RTL_R32(PABCD_CNR) & (~(1 << INET_LED_GREEN_PIN))));
	RTL_W32(PABCD_CNR, (RTL_R32(PABCD_CNR) & (~(1 << INET_LED_ORANGE_PIN))));

	RTL_W32(PABCD_DIR, (RTL_R32(PABCD_DIR) | (1 << INET_LED_GREEN_PIN)));
	RTL_W32(PABCD_DIR, (RTL_R32(PABCD_DIR) | (1 << INET_LED_ORANGE_PIN)));

	//RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << INET_LED_GREEN_PIN))));
	RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << INET_LED_ORANGE_PIN))));
	RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << INET_LED_GREEN_PIN)));
	//RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << INET_LED_ORANGE_PIN)));	


	}
	
}
#endif
void autoconfig_gpio_init(void)
{
#if defined(CONFIG_RTL865X) || defined(CONFIG_RTL8196B)
	RTL_W32(PABCD_CNR,(RTL_R32(PABCD_CNR)&(~(1 << AUTOCFG_BTN_PIN))));
	RTL_W32(PABCD_CNR,(RTL_R32(PABCD_CNR)&(~(1 << AUTOCFG_LED_PIN))));
#endif

	// Set GPIOA pin 1 as input pin for auto config button
	RTL_W32(RTL_GPIO_PABDIR, (RTL_R32(RTL_GPIO_PABDIR) & (~(1 << AUTOCFG_BTN_PIN))));

	// Set GPIOA ping 3 as output pin for auto config led
	RTL_W32(RTL_GPIO_PABDIR, (RTL_R32(RTL_GPIO_PABDIR) | (1 << AUTOCFG_LED_PIN)));

	// turn off auto config led in the beginning
	RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) | (1 << AUTOCFG_LED_PIN)));
}
#ifdef CONFIG_RTL8196B_GW_MP

#if defined(CONFIG_RTL8196B_GW_MP_KLD)
	#define PS_LED_GREEN_PIN	18
	#define PS_LED_ORANGE_PIN	3
	#define INET_LED_GREEN_PIN	8
	#define INET_LED_ORANGE_PIN	17
	#define GPIO_IRQ_NUM		1	
	#define AUTOCFG_LED_PIN_MP        20
	#define RESET_LED_PIN_MP           18
#endif
void all_led_on(void)
{
	//printk("Into MP GPIO");
	 #ifdef CONFIG_RTL8196B_GW_MP_KLD
	RTL_W32(0xB8000030, (RTL_R32(0xB8000030) | 0x00F00F80 ));
	RTL_W32(PABCD_CNR, (RTL_R32(PABCD_CNR) & (~(1 << PS_LED_GREEN_PIN))));
	RTL_W32(PABCD_CNR, (RTL_R32(PABCD_CNR) & (~(1 << PS_LED_ORANGE_PIN))));

	RTL_W32(PABCD_DIR, (RTL_R32(PABCD_DIR) | (1 << PS_LED_GREEN_PIN)));
	RTL_W32(PABCD_DIR, (RTL_R32(PABCD_DIR) | (1 << PS_LED_ORANGE_PIN)));
	
	RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << PS_LED_GREEN_PIN)));
	RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << PS_LED_ORANGE_PIN))));

	/* inet_led init setting */
	RTL_W32(PABCD_CNR, (RTL_R32(PABCD_CNR) & (~(1 << INET_LED_GREEN_PIN))));
	RTL_W32(PABCD_CNR, (RTL_R32(PABCD_CNR) & (~(1 << INET_LED_ORANGE_PIN))));

	RTL_W32(PABCD_DIR, (RTL_R32(PABCD_DIR) | (1 << INET_LED_GREEN_PIN)));
	RTL_W32(PABCD_DIR, (RTL_R32(PABCD_DIR) | (1 << INET_LED_ORANGE_PIN)));
	
	RTL_W32(PABCD_CNR, (RTL_R32(PABCD_CNR) & (~(1 << AUTOCFG_LED_PIN_MP))));

	RTL_W32(PABCD_DIR, (RTL_R32(PABCD_DIR) | (1 << AUTOCFG_LED_PIN_MP)));

	//RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << INET_LED_GREEN_PIN))));
	//RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << INET_LED_ORANGE_PIN)));
	RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << PS_LED_GREEN_PIN))));	
	RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << PS_LED_ORANGE_PIN))));
	RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << INET_LED_GREEN_PIN))));
	RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << INET_LED_ORANGE_PIN))));
	RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << AUTOCFG_LED_PIN_MP))));
	RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << RESET_LED_PIN_MP))));
	#endif

}
#endif
void autoconfig_gpio_off(void)
{
	RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) | (1 << AUTOCFG_LED_PIN)));
	AutoCfg_LED_Blink = 0;
}


void autoconfig_gpio_on(void)
{
	RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1 << AUTOCFG_LED_PIN))));
	AutoCfg_LED_Blink = 0;
}


#if (CONFIG_DNI_LED)
//max_on: led ON time during blink (unit: Tick)
//max_off: led OFF time during blink (unit: Tick)
void autoconfig_gpio_blink(unsigned int max_on, unsigned int max_off)
#else
void autoconfig_gpio_blink(void)
#endif
{
	RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1 << AUTOCFG_LED_PIN))));
#if  defined(CONFIG_RTL8196B_KLD)	
	// need  fast blink (0.25 sec on and 0.25 sec off)
	AutoCfg_LED_Blink = 3;
	AutoCfg_LED_Toggle = 1;
	mod_timer(&wps_QuickBlink_timer, jiffies + 25);
#else
	AutoCfg_LED_Blink = 1;
	AutoCfg_LED_Toggle = 1;
#endif
#if defined(CONFIG_RTL865X_SC)
	ResetToAutoCfgBtn = 0;
#endif
#if (CONFIG_DNI_LED)
	//while LED_ON or LED_off hit their max_count, toggle to the other state(ON->OFF/OFF->ON)
	AutoCfg_LED_ON_MAx_Count = max_on / RTL_GPIO_TIMER_RESOLUTION;
	AutoCfg_LED_OFF_MAx_Count = max_off / RTL_GPIO_TIMER_RESOLUTION;
	AutoCfg_LED_ON_Count = AutoCfg_LED_OFF_Count = 0;
#if 0
	AutoCfg_LED_Blink_Timeout = (timeout * TICKS_PER_SEC / RTL_GPIO_TIMER_RESOLUTION);
	AutoCfg_LED_Blink_Time = 0;
#endif
#endif
}


#if defined(CONFIG_RTL8186_TR) || defined(CONFIG_RTL8196B_TR)
void autoconfig_gpio_blink_quick(void)
{
	//Brad Modify for tr 11n
	//RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1 << AUTOCFG_LED_PIN))));
	RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1 << RESET_LED_PIN))));

	AutoCfg_LED_Blink = 2;
	AutoCfg_LED_Toggle = 1;
}
#endif

#ifdef CONFIG_RTL865X_AC
void autoconfig_gpio_blink_quick(void)
{
	//Brad Modify for tr 11n
	//RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1 << AUTOCFG_LED_PIN))));
	RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1 << RESET_LED_PIN))));

	AutoCfg_LED_Blink = 2;
	AutoCfg_LED_Toggle = 1;
}
#endif
#endif // AUTO_CONFIG

#if defined(CONFIG_DNI_LED)
static unsigned int		PW_LED_Blink;
static unsigned int		PW_LED_Toggle;
static unsigned int		PW_LED_ON_MAx_Count;
static unsigned int		PW_LED_OFF_MAx_Count;
static unsigned int		PW_LED_ON_Count;
static unsigned int		PW_LED_OFF_Count;



void pw_led_init(void)
{
	RTL_W32(RTL_GPIO_PABCNR, (RTL_R32(RTL_GPIO_PABCNR) & (~(1 << DNI_POWER_LED))));
	RTL_W32(RTL_GPIO_PABDIR, (RTL_R32(RTL_GPIO_PABDIR) | ((1 << DNI_POWER_LED))));
}


void pw_led_off(void)
{
	RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) | ((1 << DNI_POWER_LED))));
	PW_LED_Blink = 0;
}


void pw_led_on(void)
{
	RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1 << DNI_POWER_LED))));
	PW_LED_Blink = 0;
}


//note that params on_max and off_max must be (N * RTL_GPIO_TIMER_RESOLUTION), N >= 1
void pw_led_blink(unsigned int on_max, unsigned int off_max)  //in Ticks
//void pw_led_blink()
{
	RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1 << DNI_POWER_LED))));
	PW_LED_Blink = 1;
	PW_LED_Toggle = 1;
	//while LED_ON or LED_off hit their max_count, toggle to the other state(ON->OFF/OFF->ON)
	PW_LED_ON_MAx_Count = on_max / RTL_GPIO_TIMER_RESOLUTION;
	PW_LED_OFF_MAx_Count = off_max / RTL_GPIO_TIMER_RESOLUTION;
	PW_LED_ON_Count = PW_LED_OFF_Count = 0;
}
#endif  //#if defined(CONFIG_DNI_LED)


#if 0
static int reset_flash_default(void *data)
{
	char *argv[3], *envp[1] = {NULL};
	int i = 0;
	int reset_default=(int)data;

	argv[i++] = "/bin/flash";
	argv[i++] = "reset";
	argv[i] = NULL;

	if(reset_default)
	{
		printk("Going to Reload Default\n");
		if (call_usermodehelper(argv[0], argv, envp,1))
			printk("failed to Reset to default\n");
	}
	machine_restart(0);
	return 0;
}
#endif

//dvd.chen, add for sending reset_default signal to user space init task.
#define DNI_RESET_DEFAULT_NOTIFY
#ifdef DNI_RESET_DEFAULT_NOTIFY
#include <linux/sched.h>
#define SIG_RST_DFT	71  //this must be the same as the one in init program
void send_rst_dft_sig()
{
	struct task_struct *initd;

	printk("##sending SIG_RST_DFT to init task\n");
	initd = find_task_by_pid(1); /* init pid is always 1 */	
	send_sig(SIG_RST_DFT, initd, 0); /* send signal to PID 1 process */
}
#endif  //#ifdef DNI_LINK_CHANGE_NOTIFY


static void rtl_gpio_timer(unsigned long data)
{
	unsigned int pressed=1;
#if defined(CONFIG_DNI_LED)
	static int pw_led_blink_triggered = 0;
#endif

#if (defined(CONFIG_RTL8186_TR) || defined(CONFIG_RTL8196B_TR)) && defined(AUTO_CONFIG)
	static int cnt=0;

	if (AutoCfg_LED_Blink == 2)
	{
		if (AutoCfg_LED_Toggle) {
			RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) | (1 << AUTOCFG_LED_PIN)));
		}
		else {
			RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1 << AUTOCFG_LED_PIN))));
		}
		AutoCfg_LED_Toggle = AutoCfg_LED_Toggle? 0 : 1;
	}

	if (cnt++%4)
		goto ret_timer;
#endif

#if ( defined(CONFIG_RTL865X_AC) || defined(CONFIG_RTL865X_KLD) || defined(CONFIG_RTL8196B_KLD)) && defined(AUTO_CONFIG)
	static int cnt=0;

	if (AutoCfg_LED_Blink == 2)
	{
		if (AutoCfg_LED_Toggle) {
			RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) | (1 << AUTOCFG_LED_PIN)));
		}
		else {
			RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1 << AUTOCFG_LED_PIN))));
		}
		AutoCfg_LED_Toggle = AutoCfg_LED_Toggle? 0 : 1;
	}

	if (cnt++%4)
		goto ret_timer;
#endif
#ifdef CONFIG_RTL8196B_KLD
	if(test_mode_flag == '1')
	{
		if(RTL_R32(RTL_GPIO_PABDATA) & (1 << AUTOCFG_BTN_PIN))
		{
			wps_btn_flag = 0;
		}
		else
		{
			wps_btn_flag = 1;
		}
	}
#endif	
#ifdef CONFIG_RTL_EB8186
	if (RTL_R32(RTL_GPIO_PABDATA) & (1 << RESET_BTN_PIN))
#endif
#if defined(CONFIG_RTL865X)
	if (RTL_R32(PEFGH_DAT) & (1 << RESET_BTN_PIN))
#elif defined(CONFIG_RTL8196B)	
	if (RTL_R32(PABCD_DAT) & (1 << RESET_BTN_PIN))
#endif
	{
		pressed = 0;
#if defined(CONFIG_RTL8196B_KLD)			
		rst_btn_flag = 0;
#endif
#if !defined(CONFIG_RTL8186_TR) && !defined(CONFIG_RTL8196B_TR) && !defined(CONFIG_RTL865X_AC) && !defined(CONFIG_RTL865X_KLD) && !defined(CONFIG_RTL8196B_KLD)
		//turn off LED0
		#ifndef CONFIG_RTL8196B_GW_MP
		RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) | ((1 << RESET_LED_PIN))));
		#endif
#endif
	}
	else
	{
#if defined(CONFIG_RTL8196B_KLD)	
		rst_btn_flag = 1;
#endif
#if defined(CONFIG_DNI_LED)
			if ( (probe_counter+1) % (TICKS_PER_SEC/RTL_GPIO_TIMER_RESOLUTION) == 0 )
			{
				DPRINTK("Key pressed %d!\n", (probe_counter+1)/(TICKS_PER_SEC/RTL_GPIO_TIMER_RESOLUTION));
			}
#else
			DPRINTK("Key pressed %d!\n", probe_counter+1);
#endif
	}

	if (probe_state == PROBE_NULL)
	{
		if (pressed)
		{
			probe_state = PROBE_ACTIVE;
			probe_counter++;
		}
		else
			probe_counter = 0;
	}
	else if (probe_state == PROBE_ACTIVE)
	{
		if (pressed)
		{
			probe_counter++;
			
#if defined(CONFIG_DNI_LED)
			//until pressed for PROBE_TIME, 
			if( (probe_counter >=  PROBE_TIME) && (!pw_led_blink_triggered) )
			{
				pw_led_blink(30, 70);
				pw_led_blink_triggered = 1;
			}
#endif

#ifdef DVD_TAKE_OFF
#if defined(CONFIG_RTL8186_TR) || defined(CONFIG_RTL8196B_TR)
			if ((probe_counter >=1 ) && (probe_counter <=PROBE_TIME))
#elif defined(CONFIG_RTL865X_AC) || defined(CONFIG_RTL865X_KLD) || defined(CONFIG_RTL8196B_KLD)
			if ((probe_counter >=1 ) && (probe_counter <=PROBE_TIME))
#else
	#if defined(CONFIG_DNI_LED)
			if ((probe_counter >=REBOOT_TIME ) && (probe_counter <=PROBE_TIME))	
	#else
			if ((probe_counter >=2 ) && (probe_counter <=PROBE_TIME))
	#endif
#endif
			{
#if 0
//#ifndef CONFIG_RTL_EB8186
				led0enable=1;
				// turn on LED0
				RTL_W8(RTL_WLAN_CR9346, 0xc0);
				RTL_W8(RTL_WLAN_CONFIG0, 0x10);
				RTL_W8(RTL_WLAN_PSR,(RTL_R8(RTL_WLAN_PSR)&0xEF));
#else
#if !defined(CONFIG_RTL8186_TR) && !defined(CONFIG_RTL8196B_TR) && !defined(CONFIG_RTL865X_AC) && !defined(CONFIG_RTL865X_KLD) && !defined(CONFIG_RTL8196B_KLD)
				DPRINTK("2-5 turn on led\n");
				//turn on LED0
				RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1 << RESET_LED_PIN))));
#endif
#endif
			}
			else if (probe_counter >= PROBE_TIME)
			{
#if 0
//#ifndef CONFIG_RTL_EB8186
				led0enable=1;
				RTL_W8(RTL_WLAN_CR9346, 0xc0);
				RTL_W8(RTL_WLAN_CONFIG0, 0x10);
				// sparkling LED0
				if (probe_counter & 1)
					RTL_W8(RTL_WLAN_PSR, (RTL_R8(RTL_WLAN_PSR)|0x10));
				else
					RTL_W8(RTL_WLAN_PSR, (RTL_R8(RTL_WLAN_PSR)&0xEF));
#else
				// sparkling LED0
				DPRINTK(">5 \n");

#ifdef CONFIG_RTL8186_KB
				disable_power_led_blink=0;
#elif !defined(CONFIG_RTL8186_TR) && !defined(CONFIG_RTL8196B_TR) && !defined(CONFIG_RTL865X_AC) && !defined(CONFIG_RTL865X_KLD) && !defined(CONFIG_RTL8196B_KLD)
				if (probe_counter & 1)
					RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) | ((1 << RESET_LED_PIN))));
				else
					RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1 << RESET_LED_PIN))));
#endif // CONFIG_RTL8186_KB
#endif // CONFIG_RTL_EB8186
			}
#endif  //DVD_TAKE_OFF
		}
		else
		{
#if defined (CONFIG_DNI_LED)
			pw_led_blink_triggered = 0;
			pw_led_on();
#endif
#ifdef DVD_TAKE_OFF		
			#if defined(CONFIG_RTL865X_SC)
			if (probe_counter < 5)
			#else
#if defined (CONFIG_DNI_LED)
			pw_led_on();
			if (probe_counter < REBOOT_TIME)
#else
			if (probe_counter < 2)
#endif
			#endif
			{
				probe_state = PROBE_NULL;
				probe_counter = 0;
				DPRINTK("<2 \n");
				#if defined(CONFIG_RTL865X_SC)
				ResetToAutoCfgBtn = 1;
				#endif
			}
			else
#endif  //DVD_TAKE_OFF
			if (probe_counter >= PROBE_TIME)
			{
#if defined(CONFIG_RTL8196B_KLD)			
				if(test_mode_flag == '1')
				{					
				}
				else
				{
#endif				
#if 0
//#ifndef CONFIG_RTL_EB8186
				led0enable=1;
#endif

#ifdef CONFIG_RTL8186_KB
				disable_power_led_blink=0;
#endif

				//reload default
				default_flag = '1';
//added by dvd.chen
#ifdef DNI_RESET_DEFAULT_NOTIFY
				//send signal to init task
				send_rst_dft_sig();		
#endif

#ifdef CONFIG_RTL865X_KLD
			if(kld_model==1){
				ps_led_flag = '0'; 
				ps_led_process(); 
			}
#endif
#ifdef CONFIG_RTL8196B_KLD
			if(kld_model==1){
				ps_led_flag = '0'; 
				ps_led_process(); 
			}
#endif

				//kernel_thread(reset_flash_default, (void *)1, SIGCHLD);
				return;
#if defined(CONFIG_RTL8196B_KLD)			
				}
#endif	

			}
			else
			{
#if defined(CONFIG_RTL8196B_KLD)			
				if(test_mode_flag == '1')
				{					
				}
				else
				{
#endif	
#if 0
//#ifndef CONFIG_RTL_EB8186
				RTL_W8(RTL_WLAN_PSR , (RTL_R8(RTL_WLAN_PSR)|0x10));
				RTL_W8(RTL_WLAN_CONFIG0, (RTL_R8(RTL_WLAN_CONFIG0)&(~0x10)));
#endif
				DPRINTK("2-5 reset\n");
#ifdef CONFIG_RTL865X_KLD
			if(kld_model==1){
				ps_led_flag = '0';//brad
				ps_led_process();//brad
			}
#endif
#ifdef CONFIG_RTL8196B_KLD
			if(kld_model==1){
				ps_led_flag = '0';//brad
				ps_led_process();//brad
			}
#endif

				kill_proc(1,SIGTERM,1);
				//kernel_thread(reset_flash_default, 0, SIGCHLD);
				return;
#if defined(CONFIG_RTL8196B_KLD)			
				}
#endif
			}
		}
	}

#ifdef AUTO_CONFIG
	if (AutoCfg_LED_Blink==1)
	{
#if defined(CONFIG_DNI_LED)
		if (AutoCfg_LED_Toggle) {
			if( AutoCfg_LED_OFF_Count == 0) {
				DPRINTK("wps off\n");
				RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) | (1 << AUTOCFG_LED_PIN)));
			}
			if( (++AutoCfg_LED_OFF_Count) >= AutoCfg_LED_OFF_MAx_Count ) {
				AutoCfg_LED_OFF_Count = 0;
				AutoCfg_LED_Toggle = 0;
			}
		}
		else {
			if( AutoCfg_LED_ON_Count == 0) {
				DPRINTK("wps on\n");
				RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1 << AUTOCFG_LED_PIN))));
			}
			if( (++AutoCfg_LED_ON_Count) >= AutoCfg_LED_ON_MAx_Count ) {
				AutoCfg_LED_ON_Count = 0;
				AutoCfg_LED_Toggle = 1;
			}
		}
	#if 0
		//check if having blink timeout
		if( AutoCfg_LED_Blink_Timeout ) {
			if(++AutoCfg_LED_Blink_Time >= AutoCfg_LED_Blink_Timeout)
				AutoCfg_LED_Blink = 0;
		}
	#endif
#else
		if (AutoCfg_LED_Toggle) {
			RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) | (1 << AUTOCFG_LED_PIN)));
		}
		else {
			RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1 << AUTOCFG_LED_PIN))));
		}
		AutoCfg_LED_Toggle = AutoCfg_LED_Toggle? 0 : 1;
#endif
	}
#endif

#if defined(CONFIG_DNI_LED)
	if (PW_LED_Blink==1)
	{
		//DPRINTK("[%d, %d, %d]\n", PW_LED_Toggle, PW_LED_OFF_Count, PW_LED_ON_Count);
		if (PW_LED_Toggle) {
			if( PW_LED_OFF_Count == 0 ) {
				DPRINTK("pw off\n");
				RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) | (1 << DNI_POWER_LED)));
			}
			if( (++PW_LED_OFF_Count) >= PW_LED_OFF_MAx_Count ) {
				PW_LED_OFF_Count = 0;
				PW_LED_Toggle = 0;
			}
		}
		else {
			if( PW_LED_ON_Count == 0 ) {
				DPRINTK("pw on\n");
				RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1 << DNI_POWER_LED))));
			}
			if( (++PW_LED_ON_Count) >= PW_LED_ON_MAx_Count) {
				PW_LED_ON_Count = 0;
				PW_LED_Toggle = 1;
			}
		}
		
		//PW_LED_Toggle = PW_LED_Toggle? 0 : 1;
	}
#endif

#if (defined(CONFIG_RTL8186_TR) || defined(CONFIG_RTL8196B_TR)) && defined(AUTO_CONFIG)
ret_timer:
	mod_timer(&probe_timer, jiffies + 25);
#elif defined(CONFIG_RTL865X_AC) && defined(AUTO_CONFIG)
ret_timer:
	mod_timer(&probe_timer, jiffies + 25);
#elif defined(CONFIG_RTL865X_KLD) && defined(AUTO_CONFIG)
ret_timer:
	mod_timer(&probe_timer, jiffies + 25);
#elif defined(CONFIG_RTL8196B_KLD) && defined(AUTO_CONFIG)
ret_timer:
	mod_timer(&probe_timer, jiffies + 25);	
#elif defined(CONFIG_DNI_LED)
	mod_timer(&probe_timer, jiffies + RTL_GPIO_TIMER_RESOLUTION);
#else
	mod_timer(&probe_timer, jiffies + 100);
#endif
}

#ifdef AUTO_CONFIG
#if defined(CONFIG_RTL865X_KLD) || defined(CONFIG_RTL8196B_KLD) || defined(DET_WPS_SPEC)
static void gpio_interrupt_isr(int irq, void *dev_instance, struct pt_regs *regs)
{
	wps_button_push = 1;   	  	
  	RTL_W32(PABCD_ISR, RTL_R32(PABCD_ISR)); 	
}
#endif

static int read_proc(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	int len;
	char flag;

#if defined(CONFIG_RTL865X_KLD) || defined(CONFIG_RTL8196B_KLD)
	if (wps_button_push) {
		flag = '1';
		wps_button_push = 0;		
	}
	else
		flag = '0';
		
#elif defined(DET_WPS_SPEC)
// 2009-0414		
	if (wps_button_push) {
		flag = '1';
		wps_button_push = 0;		
	}
	else{
		if (RTL_R32(RTL_GPIO_PABDATA) & (1 << AUTOCFG_BTN_PIN)){
			flag = '0';
		}else{
			//panic_printk("wps button be held \n");
			flag = '1';
		}

	}
// 2009-0414		
#else
#if defined(CONFIG_RTL865X_SC)
	if ((RTL_R32(RTL_GPIO_PABDATA) & (1 << AUTOCFG_BTN_PIN)) && ResetToAutoCfgBtn == 0)
#else
	if (RTL_R32(RTL_GPIO_PABDATA) & (1 << AUTOCFG_BTN_PIN))
#endif
		flag = '0';
	else
		flag = '1';
#endif // CONFIG_RTL865X_KLD		
		
	len = sprintf(page, "%c\n", flag);


	if (len <= off+count) *eof = 1;
	*start = page + off;
	len -= off;
	if (len > count) len = count;
	if (len < 0) len = 0;
	return len;
}

#if defined(CONFIG_RTL8186_TR) || defined(CONFIG_RTL8196B_TR)
static int read_wps_led_proc(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	int len;
	char flag;

	if (RTL_R32(RTL_GPIO_PABDATA) & (1 << AUTOCFG_LED_PIN))
		flag = '0';
	else
		flag = '1';
	len = sprintf(page, "%c\n", flag);

	if (len <= off+count) *eof = 1;
	*start = page + off;
	len -= off;
	if (len > count) len = count;
	if (len < 0) len = 0;
	return len;
}
#endif

#if defined(CONFIG_RTL865X_AC) || defined(CONFIG_RTL865X_KLD) || defined(CONFIG_RTL8196B_KLD)
static int read_wps_led_proc(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	int len;
	char flag;

	if (RTL_R32(RTL_GPIO_PABDATA) & (1 << AUTOCFG_LED_PIN))
		flag = '0';
	else
		flag = '1';
	len = sprintf(page, "%c\n", flag);

	if (len <= off+count) *eof = 1;
	*start = page + off;
	len -= off;
	if (len > count) len = count;
	if (len < 0) len = 0;
	return len;
}
#ifdef CONFIG_RTL8196B_KLD
static int read_rst_btn_proc(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	int len;
	char flag;

	if (rst_btn_flag == 0)
		flag = '0';
	else
		flag = '1';
	len = sprintf(page, "%c\n", flag);

	if (len <= off+count) *eof = 1;
	*start = page + off;
	len -= off;
	if (len > count) len = count;
	if (len < 0) len = 0;
	return len;
}

static int read_wps_btn_proc(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	int len;
	char flag;

	if (wps_btn_flag == 0)
		flag = '0';
	else
		flag = '1';
	len = sprintf(page, "%c\n", flag);

	if (len <= off+count) *eof = 1;
	*start = page + off;
	len -= off;
	if (len > count) len = count;
	if (len < 0) len = 0;
	return len;
}
#endif
#endif

#ifdef CONFIG_RTL_KERNEL_MIPS16_CHAR
__NOMIPS16
#endif 
static int write_proc(struct file *file, const char *buffer,
				unsigned long count, void *data)
{
	char flag[20];
//Brad add for update flash check 20080711
#if !defined(CONFIG_RTL8186_TR) && !defined(CONFIG_RTL8196B_TR) && !defined(CONFIG_RTL865X_AC) && !defined(CONFIG_RTL865X_KLD) && !defined(CONFIG_RTL8196B_KLD) && !defined(CONFIG_RTL865X_SC)
	char start_count[10], wait[10];
#endif
	if (count < 2)
		return -EFAULT;

	DPRINTK("file: %08x, buffer: %s, count: %lu, data: %08x\n",
		(unsigned int)file, buffer, count, (unsigned int)data);

	if (buffer && !copy_from_user(&flag, buffer, 1)) {
#if defined(CONFIG_RTL8196B_KLD)		
		if(test_mode_flag == '1' && flag[0] != '0' && flag[0] != '9' )
		{
			return count;
		}
#endif		
		if (flag[0] == 'E')
			autoconfig_gpio_init();
#if defined(CONFIG_RTL8196B_KLD)	
		else if (flag[0] == '9' && test_mode_flag == '1')
			autoconfig_gpio_on();
#endif		
		else if (flag[0] == '0')
			autoconfig_gpio_off();
		else if (flag[0] == '1')
			autoconfig_gpio_on();
#if (CONFIG_DNI_LED)
		else if (flag[0] == '2')  //normal blink rate(0.5/0.5) to indicate WPS start
			autoconfig_gpio_blink(50, 50);
		else if (flag[0] == '3')  //fast blink rate (0.1/0.1) to indicate WPS error
			autoconfig_gpio_blink(10, 10);
#else
		else if (flag[0] == '2')
			autoconfig_gpio_blink();
#endif
#ifdef CONFIG_RTL8196B_GW_MP
		else if (flag[0] == '9') // disable system led
                {
			all_led_on();
		}
#endif	
//Brad add for update flash check 20080711
#if !defined(CONFIG_RTL8186_TR) && !defined(CONFIG_RTL8196B_TR) && !defined(CONFIG_RTL865X_AC) && !defined(CONFIG_RTL865X_KLD) && !defined(CONFIG_RTL8196B_KLD) && !defined(CONFIG_RTL865X_SC)
		else if (flag[0] == '4'){
			start_count_time= 1;
			sscanf(buffer, "%s %s", start_count, wait);
			Reboot_Wait = (simple_strtol(wait,NULL,0))*100;
		}
#endif
#if defined(CONFIG_RTL8186_TR) || defined(CONFIG_RTL8196B_TR)
		else if (flag[0] == '3') // update flash in special case
			flash_write_flag = 0x8000;
		else if (flag[0] == '4') // enable system led blinking
			system_led_blink = 1;
		else if (flag[0] == '5') // disable system led blinking
			system_led_blink = 0;
		else if (flag[0] == '6')
			autoconfig_gpio_blink_quick();
		else if (flag[0] == '7') // disable system led
			system_led_blink = 2;
#if 0 //Brad disable, since the check is not precise for system status			
		else if(flag[0] =='8'){ //update keep alive flag
			//printk("update keep alive flag\n");
			sys_keep_alive++;
		}
#endif
#endif

#ifdef CONFIG_RTL865X_AC
		else if (flag[0] == '3') // update flash in special case
			flash_write_flag = 0x8000;
		else if (flag[0] == '4') // enable system led blinking
			system_led_blink = 1;
		else if (flag[0] == '5') // disable system led blinking
			system_led_blink = 0;
		else if (flag[0] == '6')
			autoconfig_gpio_blink_quick();
		else if (flag[0] == '7') // disable system led
			system_led_blink = 2;
#if 0 //Brad disable, since the check is not precise for system status
		else if(flag[0] =='8'){ //update keep alive flag
			//printk("update keep alive flag\n");
			sys_keep_alive++;
		}
#endif
#endif
#ifdef CONFIG_RTL865X_KLD
		else if (flag[0] == '3') // update flash in special case
			flash_write_flag = 0x8000;
		else if (flag[0] == '4') // enable system led blinking
			system_led_blink = 1;
		else if (flag[0] == '5') // disable system led blinking
			system_led_blink = 0;
//		else if (flag[0] == '6')
//			autoconfig_gpio_blink_quick();
		else if (flag[0] == '7') // disable system led
			system_led_blink = 2;
#if 0 //Brad disable, since the check is not precise for system status
		else if(flag[0] =='8'){ //update keep alive flag
			//printk("update keep alive flag\n");
			sys_keep_alive++;
		}
#endif
#endif
#ifdef CONFIG_RTL8196B_KLD
		else if (flag[0] == '3') // update flash in special case
			flash_write_flag = 0x8000;
		else if (flag[0] == '4') // enable system led blinking
			system_led_blink = 1;
		else if (flag[0] == '5') // disable system led blinking
			system_led_blink = 0;
//		else if (flag[0] == '6')
//			autoconfig_gpio_blink_quick();
		else if (flag[0] == '7') // disable system led
			system_led_blink = 2;
#if 0 //Brad disable, since the check is not precise for system status
		else if(flag[0] =='8'){ //update keep alive flag
			//printk("update keep alive flag\n");
			sys_keep_alive++;
		}
#endif
#endif
#ifdef CONFIG_RTL8186_KB
		else if (flag[0] == '3')
			disable_power_led_blink=1;
		else if (flag[0] == '4')
			disable_power_led_blink=0;
#ifdef CONFIG_SQUASHFS
		else if (flag[0] == '5') {
			disable_power_led_blink=3; // set power-led off
#if 0
			extern int init_squashfs_fs(void);
			extern void exit_squashfs_fs(void);
			exit_squashfs_fs();
			init_squashfs_fs();
#endif
		}
#endif
#endif
		else
			{}

		return count;
	}
	else
		return -EFAULT;
}
#endif // AUTO_CONFIG

static int default_read_proc(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	int len;

	len = sprintf(page, "%c\n", default_flag);
	if (len <= off+count) *eof = 1;
	  *start = page + off;
	len -= off;
	if (len>count) len = count;
	if (len<0) len = 0;
	  return len;
}

#ifdef CONFIG_RTL_KERNEL_MIPS16_CHAR
__NOMIPS16
#endif 
static int default_write_proc(struct file *file, const char *buffer,
				unsigned long count, void *data)
{
	if (count < 2)
		return -EFAULT;
	if (buffer && !copy_from_user(&default_flag, buffer, 1)) {
		return count;
	}
	return -EFAULT;
}

#ifdef READ_RF_SWITCH_GPIO
static int rf_switch_read_proc(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	int len;
	char flag;

	if (RTL_R32(RTL_GPIO_PABDATA) & (0x00080000)){
		flag = '1';
	}else{
		flag = '0';
	}
	len = sprintf(page, "%c\n", flag);

	if (len <= off+count) *eof = 1;
	*start = page + off;
	len -= off;
	if (len > count) len = count;
	if (len < 0) len = 0;
	return len;
}
#endif

#if defined(CONFIG_DNI_LED)
#if 0
static int pw_led_read_proc(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	int len;
	char flag;


	if (RTL_R32(RTL_GPIO_PABDATA) & (1 << DNI_POWER_LED))
		flag = '0';
	else
		flag = '1';
		
	len = sprintf(page, "%c\n", flag);

	if (len <= off+count) *eof = 1;
	*start = page + off;
	len -= off;
	if (len > count) len = count;
	if (len < 0) len = 0;
	return len;
}
#endif

static int pw_led_write_proc(struct file *file, const char *buffer,
				unsigned long count, void *data)
{
	char flag[20];

	if (count < 2)
		return -EFAULT;

	DPRINTK("file: %08x, buffer: %s, count: %lu, data: %08x\n",
		(unsigned int)file, buffer, count, (unsigned int)data);

	if (buffer && !copy_from_user(&flag, buffer, 1)) {
		if (flag[0] == '0')
			pw_led_off();
		else if (flag[0] == '1')
			pw_led_on();
		else if (flag[0] == '2')
			pw_led_blink(30, 70);
		else
			{}

		return count;
	}
	else
		return -EFAULT;
}
#endif  //#if defined(CONFIG_DNI_LED)

#ifdef CONFIG_RTL865X_KLD
static void ps_led_process(void)
{
	switch (ps_led_flag)
	{
	case '0':	// solid orange
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << PS_LED_GREEN_PIN)));
		RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) & (~(1 << PS_LED_ORANGE_PIN))));
		break;
	case '1':	// green flash once then solid green
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << PS_LED_GREEN_PIN))));
		RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) | (1 << PS_LED_ORANGE_PIN)));
		ps_green_toggle = 1;
		mod_timer(&ps_led_timer, jiffies + 50);
		break;
	case '2':	// orange in low speed
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << PS_LED_GREEN_PIN)));
		RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) & (~(1 << PS_LED_ORANGE_PIN))));
		ps_orange_toggle = 1;
		mod_timer(&ps_led_timer, jiffies + 100);
		break;
	default:
		break;
	}
}

static void inet_led_process(void)
{
	switch (inet_led_flag)
	{
	case '0':	// all off
		RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) | (1 << INET_LED_GREEN_PIN)));
		RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) | (1 << INET_LED_ORANGE_PIN)));
		break;
	case '1':	// solid green
	case '5':
		RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) & (~(1 << INET_LED_GREEN_PIN))));
		RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) | (1 << INET_LED_ORANGE_PIN)));
		break;
	case '2':	// solid orange
	case '4':
		RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) | (1 << INET_LED_GREEN_PIN)));
		RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) & (~(1 << INET_LED_ORANGE_PIN))));
		break;
	case '3':	// green in high speed
		RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) & (~(1 << INET_LED_GREEN_PIN))));
		RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) | (1 << INET_LED_ORANGE_PIN)));
		inet_green_toggle = 1;
		mod_timer(&inet_led_timer, jiffies + 25);
		break;
	case '6':	// orange in low speed
		RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) | (1 << INET_LED_GREEN_PIN)));
		RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) & (~(1 << INET_LED_ORANGE_PIN))));
		inet_orange_toggle = 1;
		mod_timer(&inet_led_timer, jiffies + 100);
		break;
	default:
		break;
	}
}

static int write_ps_led_proc(struct file *file, const char *buffer,
				unsigned long count, void *data)
{
	unsigned int reg_cnr, reg_dir;

	if (count != 2)
		return -EFAULT;
	if (buffer && !copy_from_user(&ps_led_flag, buffer, 1)) {
		//printk("ps led flag '%c'\n", ps_led_flag);
		if(ps_led_flag=='E'){
			kld_model = 0;
			ps_led_gpio_init();
		}else if(ps_led_flag=='F'){
			kld_model = 1;
			ps_led_gpio_init();
		}else
			ps_led_process();
		return count;
	}
	return -EFAULT;
}

static int write_inet_led_proc(struct file *file, const char *buffer,
				unsigned long count, void *data)
{
	unsigned int reg_cnr, reg_dir;

	if (count != 2)
		return -EFAULT;
	if (buffer && !copy_from_user(&inet_led_flag, buffer, 1)) {
		//printk("inet led flag '%c'\n", inet_led_flag);
		inet_led_process();
		return count;
	}
	return -EFAULT;
}

static void ps_led_timer_func(unsigned long data)
{
	switch (ps_led_flag)
	{
	case '1':	// green flash once then solid green
		if (ps_green_toggle) {
			RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << PS_LED_GREEN_PIN)));
			ps_green_toggle = 0;
			mod_timer(&ps_led_timer, jiffies + 50);
		}
		else
			RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << PS_LED_GREEN_PIN))));
		break;
	case '2':	// orange in low speed
		if (ps_orange_toggle) {
			RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) | (1 << PS_LED_ORANGE_PIN)));
			ps_orange_toggle = 0;
			mod_timer(&ps_led_timer, jiffies + 200);
		}
		else {
			RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) & (~(1 << PS_LED_ORANGE_PIN))));
			ps_orange_toggle = 1;
			mod_timer(&ps_led_timer, jiffies + 100);
		}
		break;
	default:
		break;
	}
}

static void inet_led_timer_func(unsigned long data)
{
	switch (inet_led_flag)
	{
	case '3':	// green in high speed
		if (inet_green_toggle) {
			RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) | (1 << INET_LED_GREEN_PIN)));
			inet_green_toggle = 0;
		}
		else {
			RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) & (~(1 << INET_LED_GREEN_PIN))));
			inet_green_toggle = 1;
		}
		mod_timer(&inet_led_timer, jiffies + 25);
		break;
	case '6':	// orange in low speed
		if (inet_orange_toggle) {
			RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) | (1 << INET_LED_ORANGE_PIN)));
			inet_orange_toggle = 0;
			mod_timer(&inet_led_timer, jiffies + 200);
		}
		else {
			RTL_W32(PEFGH_DAT, (RTL_R32(PEFGH_DAT) & (~(1 << INET_LED_ORANGE_PIN))));
			inet_orange_toggle = 1;
			mod_timer(&inet_led_timer, jiffies + 100);
		}
		break;
		default:
			break;
	}
}
#endif

#ifdef CONFIG_RTL8196B_KLD
static void ps_led_process(void)
{
	switch (ps_led_flag)
	{
	case '0':	// solid orange
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << PS_LED_GREEN_PIN)));
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << PS_LED_ORANGE_PIN))));
		break;
	case '1':	// green flash once then solid green
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << PS_LED_GREEN_PIN))));
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << PS_LED_ORANGE_PIN)));
		ps_green_toggle = 1;
		mod_timer(&ps_led_timer, jiffies + 50);
		break;
	case '2':	// orange in low speed
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << PS_LED_GREEN_PIN)));
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << PS_LED_ORANGE_PIN))));
		ps_orange_toggle = 1;
		mod_timer(&ps_led_timer, jiffies + 100);
		break;
#ifdef CONFIG_RTL8196B_KLD
	case '3':	// solid green
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << PS_LED_GREEN_PIN))));
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << PS_LED_ORANGE_PIN)));
		break;
	case '4':	// all turn off
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << PS_LED_GREEN_PIN)));
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << PS_LED_ORANGE_PIN)));
		break;
#endif
	default:
		break;
	}
}

static void inet_led_process(void)
{
	switch (inet_led_flag)
	{
	case '0':	// all off
	case '5':
#ifdef CONFIG_RTL8196B_KLD
		if(test_mode_flag == '1')
			break;
#endif
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << INET_LED_GREEN_PIN)));
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << INET_LED_ORANGE_PIN)));
		break;
	case '1':	// solid green
#ifdef CONFIG_RTL8196B_KLD
		if(test_mode_flag == '1')
			break;
#endif
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << INET_LED_GREEN_PIN))));
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << INET_LED_ORANGE_PIN)));
		break;
	case '2':	// solid orange
	case '4':
#ifdef CONFIG_RTL8196B_KLD
		if(test_mode_flag == '1')
			break;
#endif
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << INET_LED_GREEN_PIN)));
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << INET_LED_ORANGE_PIN))));
		break;
	case '3':	// green in high speed
#ifdef CONFIG_RTL8196B_KLD
		if(test_mode_flag == '1')
			break;
#endif
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << INET_LED_GREEN_PIN))));
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << INET_LED_ORANGE_PIN)));
		inet_green_toggle = 1;
		mod_timer(&inet_led_timer, jiffies + 25);
		break;
	case '6':	// orange in low speed
#ifdef CONFIG_RTL8196B_KLD
		if(test_mode_flag == '1')
			break;
#endif
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << INET_LED_GREEN_PIN)));
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << INET_LED_ORANGE_PIN))));
		inet_orange_toggle = 1;
		mod_timer(&inet_led_timer, jiffies + 100);
		break;
#ifdef CONFIG_RTL8196B_KLD
	case '7':	// all off
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << INET_LED_GREEN_PIN)));
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << INET_LED_ORANGE_PIN)));
		break;
	case '8':	// solid green
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << INET_LED_GREEN_PIN))));
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << INET_LED_ORANGE_PIN)));
		break;
	case '9':	// solid orange
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << INET_LED_GREEN_PIN)));
		RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << INET_LED_ORANGE_PIN))));
		break;
#endif
	default:
		break;
	}
}

#ifdef CONFIG_RTL8196B_KLD
static int write_test_mode_proc(struct file *file, const char *buffer,
				unsigned long count, void *data)
{
	unsigned int reg_cnr, reg_dir;

	if (count != 2)
		return -EFAULT;
	if (buffer && !copy_from_user(&test_mode_flag, buffer, 1)) {
		
	}
	return -EFAULT;
}
#endif
static int write_ps_led_proc(struct file *file, const char *buffer,
				unsigned long count, void *data)
{
	unsigned int reg_cnr, reg_dir;

	if (count != 2)
		return -EFAULT;
	if (buffer && !copy_from_user(&ps_led_flag, buffer, 1)) {
		//printk("ps led flag '%c'\n", ps_led_flag);
		if(ps_led_flag=='E'){
			kld_model = 0;
			ps_led_gpio_init();
		}else if(ps_led_flag=='F'){
			kld_model = 1;
			ps_led_gpio_init();
		}else
			ps_led_process();
		return count;
	}
	return -EFAULT;
}

static int write_inet_led_proc(struct file *file, const char *buffer,
				unsigned long count, void *data)
{
	unsigned int reg_cnr, reg_dir;

	if (count != 2)
		return -EFAULT;
	if (buffer && !copy_from_user(&inet_led_flag, buffer, 1)) {
		//printk("inet led flag '%c'\n", inet_led_flag);
		inet_led_process();
		return count;
	}
	return -EFAULT;
}

static void ps_led_timer_func(unsigned long data)
{
	switch (ps_led_flag)
	{
	case '1':	// green flash once then solid green
		if (ps_green_toggle) {
			RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << PS_LED_GREEN_PIN)));
			ps_green_toggle = 0;
			mod_timer(&ps_led_timer, jiffies + 50);
		}
		else
			RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << PS_LED_GREEN_PIN))));
		break;
	case '2':	// orange in low speed
		if (ps_orange_toggle) {
			RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << PS_LED_ORANGE_PIN)));
			ps_orange_toggle = 0;
			mod_timer(&ps_led_timer, jiffies + 200);
		}
		else {
			RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << PS_LED_ORANGE_PIN))));
			ps_orange_toggle = 1;
			mod_timer(&ps_led_timer, jiffies + 100);
		}
		break;
	default:
		break;
	}
}

static void wps_QuickBlink_func(unsigned long data)
{

	if(AutoCfg_LED_Blink == 3){
		if (AutoCfg_LED_Toggle) {
			RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) | (1 << AUTOCFG_LED_PIN)));		
			AutoCfg_LED_Toggle = 0;
		}
		else {
			RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1 << AUTOCFG_LED_PIN))));				
			AutoCfg_LED_Toggle = 1;
		}
		mod_timer(&wps_QuickBlink_timer, jiffies + 25);	
	}
}
static void inet_led_timer_func(unsigned long data)
{
	switch (inet_led_flag)
	{
	case '3':	// green in high speed
		if (inet_green_toggle) {
			RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << INET_LED_GREEN_PIN)));
			inet_green_toggle = 0;
		}
		else {
			RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << INET_LED_GREEN_PIN))));
			inet_green_toggle = 1;
		}
		mod_timer(&inet_led_timer, jiffies + 25);
		break;
	case '6':	// orange in low speed
		if (inet_orange_toggle) {
			RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) | (1 << INET_LED_ORANGE_PIN)));
			inet_orange_toggle = 0;
			mod_timer(&inet_led_timer, jiffies + 200);
		}
		else {
			RTL_W32(PABCD_DAT, (RTL_R32(PABCD_DAT) & (~(1 << INET_LED_ORANGE_PIN))));
			inet_orange_toggle = 1;
			mod_timer(&inet_led_timer, jiffies + 100);
		}
		break;
		default:
			break;
	}
}
#endif //CONFIG_RTL8196B_KLD

int __init rtl_gpio_init(void)
{
	struct proc_dir_entry *res=NULL;

	printk("Realtek GPIO Driver for Flash Reload Default\n");

	// Set GPIOA pin 10(8181)/0(8186) as input pin for reset button
#ifdef CONFIG_RTL_EB8186
	RTL_W32(RTL_GPIO_PABDIR, (RTL_R32(RTL_GPIO_PABDIR) & (~(1 << RESET_BTN_PIN))));
#endif

#if defined(CONFIG_RTL865X)
	RTL_W32(PEFGH_CNR, (RTL_R32(PEFGH_CNR) & (~(1 << RESET_BTN_PIN))));
	RTL_W32(PEFGH_DIR, (RTL_R32(PEFGH_DIR) & (~(1 << RESET_BTN_PIN))));

#elif defined(CONFIG_RTL8196B)
#if defined(CONFIG_DNI_LED)
	//0x0fc00380
	//for testing take off scan mode
#if defined(CONFIG_DNI_JTAG_ON)
	RTL_W32(RTL_GPIO_MUX, 0x0C300080); //for GPIOC4,C1 , Switch LED-SW_P2 and GPIOA[4:0]
#else
	RTL_W32(RTL_GPIO_MUX, 0x0C300380); //for GPIOC4,C1 , Switch LED-SW_P2 and GPIOA[4:0]
#endif
	//RTL_W32(RTL_GPIO_MUX, (RTL_R32(RTL_GPIO_MUX) | (RTL_GPIO_MUX_DATA)));
	//RTL_W32(RTL_GPIO_MUX, 0x0fc00380);
#else
	RTL_W32(RTL_GPIO_MUX, (RTL_R32(RTL_GPIO_MUX) | (RTL_GPIO_MUX_DATA)));
#endif
	RTL_W32(PABCD_CNR, (RTL_R32(PABCD_CNR) & (~(1 << RESET_BTN_PIN))));
	RTL_W32(PABCD_DIR, (RTL_R32(PABCD_DIR) & (~(1 << RESET_BTN_PIN))));

#if defined(READ_RF_SWITCH_GPIO)
	RTL_W32(RTL_GPIO_PABCNR, (RTL_R32(RTL_GPIO_PABCNR) & ( ~(1<<RTL_GPIO_WIFI_ONOFF))));
	RTL_W32(RTL_GPIO_PABDIR, (RTL_R32(RTL_GPIO_PABDIR) & (~(1<<RTL_GPIO_WIFI_ONOFF))));
	RTL_W32(RTL_GPIO_PABDATA, (RTL_R32(RTL_GPIO_PABDATA) & (~(1<<RTL_GPIO_WIFI_ONOFF))));

#endif // #if defined(READ_RF_SWITCH_GPIO)
#endif // #if defined(CONFIG_RTL865X)

#if !defined(CONFIG_RTL865X_KLD) && !defined(CONFIG_RTL8196B_KLD)
	// Set GPIOA ping 2 as output pin for reset led
	RTL_W32(RTL_GPIO_PABDIR, (RTL_R32(RTL_GPIO_PABDIR) | ((1 << RESET_LED_PIN))));

#endif

#if defined(CONFIG_DNI_LED)
	//set DNI power_led for output and blink it indicating system is under booting up
	pw_led_init();
	pw_led_blink(50, 50);
	
	//create proc entry for power_led
	res = create_proc_entry("pw_led", 0, NULL);
	if (res) {
		res->read_proc = NULL; //pw_led_read_proc;
		res->write_proc = pw_led_write_proc;
	}
	else {
		printk("Realtek GPIO Driver, create pw_led proc failed!\n");
	}
#endif

#ifdef AUTO_CONFIG
	res = create_proc_entry("gpio", 0, NULL);
	if (res) {
		res->read_proc = read_proc;
		res->write_proc = write_proc;
	}
	else {
		printk("Realtek GPIO Driver, create proc failed!\n");
	}

#if defined(CONFIG_RTL8186_TR) || defined(CONFIG_RTL8196B_TR)
	res = create_proc_entry("wps_led", 0, NULL);
	if (res)
		res->read_proc = read_wps_led_proc;
	else
		printk("create wps led proc failed!\n");
#endif
#ifdef CONFIG_RTL865X_AC
	res = create_proc_entry("wps_led", 0, NULL);
	if (res)
		res->read_proc = read_wps_led_proc;
	else
		printk("create wps led proc failed!\n");
#endif
#ifdef CONFIG_RTL865X_KLD
	res = create_proc_entry("wps_led", 0, NULL);
	if (res)
		res->read_proc = read_wps_led_proc;
	else
		printk("create wps led proc failed!\n");
	
	RTL_R32(PAB_IMR) |= (0x01 << AUTOCFG_BTN_PIN*2); // enable interrupt in falling-edge		
	if (request_irq(GPIO_IRQ_NUM, gpio_interrupt_isr, SA_INTERRUPT, "rtl_gpio", NULL)) {	  
		printk("request_irq(%d) error!\n", GPIO_IRQ_NUM);		
   	}	
#endif
#ifdef CONFIG_RTL8196B_KLD
	res = create_proc_entry("wps_led", 0, NULL);
	if (res)
		res->read_proc = read_wps_led_proc;
	else
		printk("create wps led proc failed!\n");
	
	RTL_R32(PAB_IMR) |= (0x01 << AUTOCFG_BTN_PIN*2); // enable interrupt in falling-edge		
	if (request_irq(GPIO_IRQ_NUM, gpio_interrupt_isr, SA_INTERRUPT, "rtl_gpio", NULL)) {	  
		printk("request_irq(%d) error!\n", GPIO_IRQ_NUM);		
   	}	
#endif
// 2009-0414		
#if defined(DET_WPS_SPEC)
	RTL_R32(PAB_IMR) |= (0x01 << AUTOCFG_BTN_PIN*2); // enable interrupt in falling-edge		
	if (request_irq(GPIO_IRQ_NUM, gpio_interrupt_isr, SA_INTERRUPT, "rtl_gpio", NULL)) {	  
		panic_printk("gpio request_irq(%d) error!\n", GPIO_IRQ_NUM);		
   	}
#endif
// 2009-0414		
#endif

	res = create_proc_entry("load_default", 0, NULL);
	if (res) {
		res->read_proc = default_read_proc;
		res->write_proc = default_write_proc;
	}

#ifdef READ_RF_SWITCH_GPIO
	res = create_proc_entry("rf_switch", 0, NULL);
	if (res) {
		res->read_proc = rf_switch_read_proc;
		res->write_proc = NULL;
	}
#endif

#ifdef CONFIG_RTL865X_KLD
	res = create_proc_entry("ps_led", 0, NULL);
	if (res)
		res->write_proc = write_ps_led_proc;
	else
		printk("create ps led proc failed!\n");

	res = create_proc_entry("inet_led", 0, NULL);
	if (res)
		res->write_proc = write_inet_led_proc;
	else
		printk("create inet led proc failed!\n");

	init_timer(&ps_led_timer);
	ps_led_timer.data = (unsigned long)NULL;
	ps_led_timer.function = &ps_led_timer_func;

	init_timer(&inet_led_timer);
	inet_led_timer.data = (unsigned long)NULL;
	inet_led_timer.function = &inet_led_timer_func;
#endif

#ifdef CONFIG_RTL8196B_KLD
	res = create_proc_entry("test_mode", 0, NULL);
	if (res)
		res->write_proc = write_test_mode_proc;
	else
		printk("create test_mode proc failed!\n");

	res = create_proc_entry("rst_btn", 0, NULL);
	if (res)
		res->read_proc = read_rst_btn_proc;
	else
		printk("create read_rst_btn_proc failed!\n");

	res = create_proc_entry("wps_btn", 0, NULL);
	if (res)
		res->read_proc = read_wps_btn_proc;
	else
		printk("create read_wps_btn_proc failed!\n");
	
	
	res = create_proc_entry("ps_led", 0, NULL);
	if (res)
		res->write_proc = write_ps_led_proc;
	else
		printk("create ps led proc failed!\n");

	res = create_proc_entry("inet_led", 0, NULL);
	if (res)
		res->write_proc = write_inet_led_proc;
	else
		printk("create inet led proc failed!\n");

	init_timer(&ps_led_timer);
	ps_led_timer.data = (unsigned long)NULL;
	ps_led_timer.function = &ps_led_timer_func;

	init_timer(&inet_led_timer);
	inet_led_timer.data = (unsigned long)NULL;
	inet_led_timer.function = &inet_led_timer_func;
	init_timer(&wps_QuickBlink_timer);
	wps_QuickBlink_timer.data = (unsigned long)NULL;
	wps_QuickBlink_timer.function = &wps_QuickBlink_func;
#endif

	init_timer(&probe_timer);
	probe_counter = 0;
	probe_state = PROBE_NULL;
#if defined(CONFIG_DNI_LED)
	probe_timer.expires = jiffies + RTL_GPIO_TIMER_RESOLUTION;
#else
	probe_timer.expires = jiffies + 100;
#endif
	probe_timer.data = (unsigned long)NULL;
	probe_timer.function = &rtl_gpio_timer;
#if defined(CONFIG_DNI_LED)
	mod_timer(&probe_timer, jiffies + RTL_GPIO_TIMER_RESOLUTION);
#else
	mod_timer(&probe_timer, jiffies + 100);
#endif

	return 0;
}


static void __exit rtl_gpio_exit(void)
{
	printk("Unload Realtek GPIO Driver \n");
	del_timer_sync(&probe_timer);
}


module_exit(rtl_gpio_exit);
module_init(rtl_gpio_init);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("GPIO driver for Reload default");
